#include "main.h"
#define EPSILON 0.001f
#define MEMORY_ARRAY_SIZE 1000000		// Beware!!!

enum Type
{
	UNDEFINED = 0,		// = 0

	NUMERIC,
	STRING,
	ARRAY,

	TYPE_COUNT			// = (number of different types)
};

class _Node
{
public:
	_Node* mParent;
	map<string, _Node*> mElements;
	
	virtual string PrintLeft()
	{
		return "";
	}

	virtual Type GetType()
	{
		return Type::UNDEFINED;
	}

	virtual void* GetData()
	{
		return NULL;
	}
};

class Node : public _Node
{
public:
	Type mType;
	string mLeftSide;
	void* mData;

	Node(_Node* parent = NULL, const string& leftside = "ROOT", Type type = Type::UNDEFINED)
	{
		mParent = parent;
		mLeftSide = leftside; 
		mType = type;
		mData = NULL;
	}

	virtual string PrintLeft()
	{
		return mLeftSide;
	}

	virtual Type GetType()
	{
		return mType;
	}

	virtual void* GetData()
	{
		return mData;
	}
};

class Array : public _Node
{
public:
	bool mIsPointer;
	vector<int> mPointer;
	vector<float> mValue;

	Array()
	{
		mIsPointer = false;
		mParent = NULL;
	}
};

vector<Array> g_mData;

void Error(const string& error)
{
	cout << "Error: " << error << endl;
	cout << "File not parsed. " << endl;
	cout << "Press return to exit..." << endl;
	cin.get();
	exit(1);
}

bool Match(const vector<string>& data, int current, const string& token)
{
	if ((unsigned int)current + 1 >= data.size())
	{
		Error("Matching after end of file");
		return false;
	}
	else
	{
		if (data[current + 1] == token)
		{
			return true;
		}
		else
		{
			return false;
		}
	}
}

bool MatchAnyOf(const vector<string>& data, int current, const vector<string>& tokens, bool fail = false)
{
	bool matched = false;
	for (const string& s : tokens)
	{
		matched |= Match(data, current, s);
	}

	if (!matched && fail)
	{
		Error(string("Invalid token at ") + boost::lexical_cast<string>(current + 1) + string(" which is ") + boost::lexical_cast<string>(data[current + 1]));
	}

	return matched;
}

vector<string> Tuple(const string& a, const string& b)
{
	vector<string> ret;
	ret.push_back(a);
	ret.push_back(b);
	return ret;
}

vector<string> Tuple(const string& a, const string& b, const string& c)
{
	vector<string> ret;
	ret.push_back(a);
	ret.push_back(b);
	ret.push_back(c);
	return ret;
}

vector<string> Tuple(const string& a, const string& b, const string& c, const string& d)
{
	vector<string> ret;
	ret.push_back(a);
	ret.push_back(b);
	ret.push_back(c);
	ret.push_back(d);
	return ret;
}

enum GeometryType
{
	GEOMPOINT = 0,
	MULTIPOINT,
	LINESTRING,
	MULTILINESTRING,
	POLYGON,
	MULTIPOLYGON,
	GEOMETRYCOLLECTION,
	FEATURE,
	FEATURECOLLECTION,
	GEOMETRY_TYPE_COUNT
};

map<string, GeometryType> GeometryTypeString = map<string, GeometryType>();

GeometryType GetGeometryType(const string& token)
{
	auto it = GeometryTypeString.find(token);
	return it->second;
}

void ValidatePoint(_Node* node)
{
	if (node->mParent == NULL)
	{
		Error("Node \"type\" declared as root node");
	}

	_Node* parent = node->mParent;

	auto it = parent->mElements.find("coordinates");
	if (it == parent->mElements.end())
	{
		Error("Node of type Point must have coordinates");
	}

	_Node* coordinates = it->second;
	if (coordinates->GetType() != Type::ARRAY)
	{
		Error("Node \"coordinates\" must contain array");
	}
	
	Node* coordData = (Node*)coordinates;
	if (coordData->mElements.size() != 1)
	{
		Error("Invalid number of arrays");
	}

	Array* data = (Array*)(coordData->mElements.begin()->second);
	if (data->mIsPointer == true)
	{
		Error("Point must contain 2 values, 0 arrays");
	}

	if (data->mValue.size() != 2)
	{
		Error("Point doesn't contain 2 numbers");
	}
}

void ValidateMultiPointLineString(_Node* node, const string& name)
{
	if (node->mParent == NULL)
	{
		Error("Node \"type\" declared as root node");
	}

	_Node* parent = node->mParent;

	auto it = parent->mElements.find("coordinates");
	if (it == parent->mElements.end())
	{
		Error("Node of type Point must have coordinates");
	}

	_Node* coordinates = it->second;
	if (coordinates->GetType() != Type::ARRAY)
	{
		Error("Node \"coordinates\" must contain array");
	}

	Node* coordData = (Node*)coordinates;
	if (coordData->mElements.size() != 1)
	{
		Error("Invalid number of arrays");
	}

	Array* data = (Array*)(coordData->mElements.begin()->second);
	if (data->mIsPointer == false)
	{
		Error(name + " must have a collection of points");
	}

	if (data->mValue.size() > 0)
	{
		Error(name + " must contain only collection of points");
	}

	if (name == "LineString")
	{
		if (data->mPointer.size() < 2)
		{
			Error(name + " must be an array of 2 or more positions");
		}
	}

	for (int i : data->mPointer)
	{
		const Array& point = g_mData[i];
		if (point.mIsPointer == true)
		{
			Error("Point must contain 2 values, 0 arrays");
		}

		if (point.mValue.size() != 2)
		{
			Error("Point doesn't contain 2 numbers");
		}
	}
}

void ValidateMultiLineString(_Node* node)
{
	if (node->mParent == NULL)
	{
		Error("Node \"type\" declared as root node");
	}

	_Node* parent = node->mParent;

	auto it = parent->mElements.find("coordinates");
	if (it == parent->mElements.end())
	{
		Error("Node of type Point must have coordinates");
	}

	_Node* coordinates = it->second;
	if (coordinates->GetType() != Type::ARRAY)
	{
		Error("Node \"coordinates\" must contain array");
	}

	Node* coordData = (Node*)coordinates;
	if (coordData->mElements.size() != 1)
	{
		Error("Invalid number of arrays");
	}

	Array* data = (Array*)(coordData->mElements.begin()->second);
	if (data->mIsPointer == false)
	{
		Error("MultiLineString must have a collection of linestrings");
	}

	if (data->mValue.size() > 0)
	{
		Error("MultiLineString must contain only collection of linestrings");
	}

	for (int i : data->mPointer)
	{
		const Array& linestring = g_mData[i];
		if (linestring.mIsPointer == false)
		{
			Error("LineString must be a collection of points");
		}

		if (linestring.mValue.size() > 0)
		{
			Error("LineString must be just a collection of points");
		}

		if (linestring.mPointer.size() < 2)
		{
			Error("LineString must be an array of 2 or more positions");
		}

		for (int j : linestring.mPointer)
		{
			const Array& point = g_mData[j];
			if (point.mIsPointer == true)
			{
				Error("Point must contain 2 values, 0 arrays");
			}

			if (point.mValue.size() != 2)
			{
				Error("Point doesn't contain 2 numbers");
			}
		}
	}
}

void ValidatePolygon(_Node* node)
{
	if (node->mParent == NULL)
	{
		Error("Node \"type\" declared as root node");
	}

	_Node* parent = node->mParent;

	auto it = parent->mElements.find("coordinates");
	if (it == parent->mElements.end())
	{
		Error("Node of type Point must have coordinates");
	}

	_Node* coordinates = it->second;
	if (coordinates->GetType() != Type::ARRAY)
	{
		Error("Node \"coordinates\" must contain array");
	}

	Node* coordData = (Node*)coordinates;
	if (coordData->mElements.size() != 1)
	{
		Error("Invalid number of arrays");
	}

	Array* data = (Array*)(coordData->mElements.begin()->second);
	if (data->mIsPointer == false)
	{
		Error("Polygon must have a collection of linestrings");
	}

	if (data->mValue.size() > 0)
	{
		Error("Polygon must contain only collection of linestrings");
	}

	for (int i : data->mPointer)
	{
		const Array& linestring = g_mData[i];
		if (linestring.mIsPointer == false)
		{
			Error("LineString (Polygon) must be a collection of points");
		}

		if (linestring.mValue.size() > 0)
		{
			Error("LineString (Polygon) must be just a collection of points");
		}

		if (linestring.mPointer.size() < 4)
		{
			Error("LineString (Polygon) must be an array of 4 or more positions");
		}

		float x, y;
		for (unsigned int j = 0; j < linestring.mPointer.size(); j++)
		{
			const Array& point = g_mData[linestring.mPointer[j]];
			if (point.mIsPointer == true)
			{
				Error("Point must contain 2 values, 0 arrays");
			}

			if (point.mValue.size() != 2)
			{
				Error("Point doesn't contain 2 numbers");
			}

			if (j == 0)
			{
				x = point.mValue[0];
				y = point.mValue[1];
			}
			else if (j == linestring.mPointer.size() - 1)
			{
				float dx = fabsf(x - point.mValue[0]);
				float dy = fabsf(y - point.mValue[1]);

				if (dx > EPSILON || dy > EPSILON)
				{
					Error("First and last point in LineString(Polygon) is not a same point");
				}
			}
		}
	}
}

void ValidateMultiPolygon(_Node* node)
{
	if (node->mParent == NULL)
	{
		Error("Node \"type\" declared as root node");
	}

	_Node* parent = node->mParent;

	auto it = parent->mElements.find("coordinates");
	if (it == parent->mElements.end())
	{
		Error("Node of type Point must have coordinates");
	}

	_Node* coordinates = it->second;
	if (coordinates->GetType() != Type::ARRAY)
	{
		Error("Node \"coordinates\" must contain array");
	}

	Node* coordData = (Node*)coordinates;
	if (coordData->mElements.size() != 1)
	{
		Error("Invalid number of arrays");
	}

	Array* data = (Array*)(coordData->mElements.begin()->second);
	if (data->mIsPointer == false)
	{
		Error("MultiPolygon must have a collection of Polygons");
	}

	if (data->mValue.size() > 0)
	{
		Error("MultiPolygon must contain only collection of Polygons");
	}

	for (int i : data->mPointer)
	{
		const Array& polygon = g_mData[i];
		if (polygon.mIsPointer == false)
		{
			Error("Polygon must have a collection of linestrings");
		}

		if (polygon.mValue.size() > 0)
		{
			Error("Polygon must contain only collection of linestrings");
		}

		for (int j : polygon.mPointer)
		{
			const Array& linestring = g_mData[j];
			if (linestring.mIsPointer == false)
			{
				Error("LineString (MultiPolygon) must be a collection of points");
			}

			if (linestring.mValue.size() > 0)
			{
				Error("LineString (MultiPolygon) must be just a collection of points");
			}

			if (linestring.mPointer.size() < 4)
			{
				Error("LineString (MultiPolygon) must be an array of 4 or more positions");
			}

			float x, y;
			for (unsigned int k = 0; k < linestring.mPointer.size(); k++)
			{
				const Array& point = g_mData[linestring.mPointer[k]];
				if (point.mIsPointer == true)
				{
					Error("Point must contain 2 values, 0 arrays");
				}

				if (point.mValue.size() != 2)
				{
					Error("Point doesn't contain 2 numbers");
				}

				if (k == 0)
				{
					x = point.mValue[0];
					y = point.mValue[1];
				}
				else if (k == linestring.mPointer.size() - 1)
				{
					float dx = fabsf(x - point.mValue[0]);
					float dy = fabsf(y - point.mValue[1]);

					if (dx > EPSILON || dy > EPSILON)
					{
						Error("First and last point in LineString(MultiPolygon) is not a same point");
					}
				}
			}
		}
	}
}

int main(int argc, char* argv[], char** envp)
{
	string filename;

	char filenametemp[65536];
    OPENFILENAME ofn;
    ZeroMemory(&ofn, sizeof(ofn));
    ofn.lStructSize = sizeof(ofn);
    ofn.hwndOwner = NULL;
    ofn.lpstrDefExt = 0;
    ofn.lpstrFile = filenametemp;
    ofn.lpstrFile[0] = '\0';
    ofn.nMaxFile = MAX_PATH;
    ofn.lpstrFilter = "Json Files (*.json)\0*.json\0";
    ofn.nFilterIndex = 0;
    ofn.lpstrInitialDir = 0;
    ofn.lpstrTitle = "Select json file to open";
    ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
    GetOpenFileName(&ofn);

    if (strlen(ofn.lpstrFile) == 0)
	{
		Error("File not selected");
	}
	else 
	{
		filename = ofn.lpstrFile;
	}

	GeometryTypeString["Point"] = GEOMPOINT;
	GeometryTypeString["MultiPoint"] = MULTIPOINT;
	GeometryTypeString["LineString"] = LINESTRING;
	GeometryTypeString["MultiLineString"] = MULTILINESTRING;
	GeometryTypeString["Polygon"] = POLYGON;
	GeometryTypeString["MultiPolygon"] = MULTIPOLYGON;
	/*GeometryTypeString["GeometryCollection"] = GEOMETRYCOLLECTION;
	GeometryTypeString["Feature"] = FEATURE;
	GeometryTypeString["FeatureCollection"] = FEATURECOLLECTION;*/

	g_mData.reserve(MEMORY_ARRAY_SIZE);

	//////////////////////////////////////////////////////////////////////////////////////////////////
	// LEXIKAL ANALYSIS

	// Initialize symbols
	set<char> symbols;
	symbols.insert(',');
	symbols.insert('[');
	symbols.insert(']');
	symbols.insert(':');
	symbols.insert('{');
	symbols.insert('}');

	// Read file into single string
	ifstream f;
	f.open(filename, ios_base::in);
	if(f.is_open() == false)
	{
		Error("Error reading file");
	}
	string line;
	string data;
	while(getline(f, line))
	{
		data += line;
	}

	// Wrap each symbol into whitespaces
	string wrapped;
	bool str = false;
	for(unsigned int i = 0; i < data.length(); i++)
	{
		char c = data[i];
		if (c == '"')
		{
			str = !str;
		}

		if (str == false)
		{
			if(symbols.find(c) == symbols.end()) // If it cant find symbol c in set symbols
			{
				if(c == '|')
				{
					wrapped += ' ';
				}
				else
				{
					wrapped += c;
				}
				continue;
			}

			wrapped += string("|") + c + "|";
		}
		else
		{
			wrapped += c;
		}
	}

	// Split on whitespaces
	vector<string> tokens;
	split(tokens, wrapped, is_any_of("|"));

	// Trim each token (remove whitespaces from left and right)
	for(string& s : tokens)
	{
		trim(s);
	}

	// Remove empty strings from tokens array
	for(vector<string>::iterator it = tokens.begin(); it != tokens.end(); )
	{
		if((*it).length() == 0)
		{
			it = tokens.erase(it);
		}
		else
		{
			++it;
		}
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////
	// SYNTACTIC ANALYSIS (PARSING)

	// Parsing...
	int level = 0;
	stack<bool> arrayLevel;
	int arrlevel = 0;
	_Node* root;
	_Node* node;
	string elementName = "";
	bool previous = false;
	stack<int> arrayId;
	bool bColonFlag = false;
	int nColonCnt = 0;

	try
	{
		for(unsigned int i = 0; i < tokens.size(); i++)
		{
			if(tokens[i] == "{")
			{
				if (level == 0)
				{
					root = new Node();
					node = root;
				}
				else if (arrlevel > 0 && arrayLevel.top() == true)
				{
					Node* temp = new Node(&(g_mData[arrayId.top()]));
					g_mData[arrayId.top()].mElements.insert(pair<string, _Node*>("node" + boost::lexical_cast<string>(g_mData[arrayId.top()].mElements.size()), temp));
					node = temp;
				}
				else
				{
					Node* temp = new Node(node, elementName);
					(*node).mElements.insert(pair<string, _Node*>(elementName, temp));
					node = temp;
				}
				arrayLevel.push(false);
				level++;
			}
			else if(tokens[i] == "}")
			{
				node = (*node).mParent;
				arrayLevel.pop();
				level--;
				if(node == NULL && level < 0)
				{
					Error("Error invalid file format");
				}
			}
			else if(tokens[i] == "[")
			{
				if (nColonCnt != 2) 
				{                            
					Error("Invalid format");  
				}   
				if (previous == true)
				{
					Node* leaf = new Node(node, elementName, Type::ARRAY);
					(*leaf).mData = new int(g_mData.size());
					(*node).mElements.insert(pair<string, Node*>(elementName, leaf));
					arrayId.push(g_mData.size());
					g_mData.push_back(Array());
					(*leaf).mElements.insert(pair<string, _Node*>("node" + boost::lexical_cast<string>((*leaf).mElements.size()), &(g_mData[arrayId.top()])));
					g_mData[arrayId.top()].mParent = leaf;
				}
				else
				{
					if (g_mData[arrayId.top()].mElements.size() == 0)
					{
						g_mData[arrayId.top()].mParent = NULL;
					}
					g_mData[arrayId.top()].mIsPointer = true;
					g_mData[arrayId.top()].mPointer.push_back(g_mData.size());
					arrayId.push(g_mData.size());
					g_mData.push_back(Array());
				}
				arrayLevel.push(true);
				arrlevel++;
			}
			else if(tokens[i] == "]")
			{
				arrlevel--;
				if (arrlevel < 0)
				{
					Error("Error invalid file format");
				}

				if (g_mData[arrayId.top()].mParent != NULL)
				{
					node = g_mData[arrayId.top()].mParent;
				}
				arrayLevel.pop();
				arrayId.pop();
			}
			else if(isTypeOf<string>(tokens[i]))
			{
				if(previous == true)
				{
					Node* leaf = new Node(node, elementName, Type::STRING);
					(*leaf).mData = new string(tokens[i].substr(1, tokens[i].length() - 2));
					(*node).mElements.insert(pair<string, Node*>(elementName, leaf));
				}
				else
				{
					elementName = tokens[i].substr(1, tokens[i].length() - 2);
				}
			}
			else if(isTypeOf<float>(tokens[i]))
			{
				if (previous == true)
				{
					Node* leaf = new Node(node, elementName, Type::NUMERIC);
					(*leaf).mData = new float(boost::lexical_cast<float>(tokens[i]));
					(*node).mElements.insert(pair<string, Node*>(elementName, leaf));
				}
				else if (arrayId.empty() != true)
				{
					if (g_mData[arrayId.top()].mIsPointer == true)
					{
						Error("Error invalid file format");
					}
					g_mData[arrayId.top()].mParent = NULL;
					g_mData[arrayId.top()].mIsPointer = false;
					g_mData[arrayId.top()].mValue.push_back(boost::lexical_cast<float>(tokens[i]));
				}
			}
			else if(tokens[i] == ":")
			{
				bColonFlag = true;
				nColonCnt++;
				previous = true;
				continue;
			}

			previous = false;
		}
	}
	catch (...)
	{
		Error("Invalid file format");
	}

	if(level != 0)
	{
		Error("Invalid file format");
	}

	if (arrlevel != 0)
	{
		Error("Error invalid file format");
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////
	// LOOP THROUGH THE CODE AND CHECK CORRECT SYNTAX

	for (unsigned int i = 0; i < tokens.size(); i++)
	{
		if (i == tokens.size() - 1)
		{
			break;
		}

		if (isTypeOf<string>(tokens[i]))
		{
			MatchAnyOf(tokens, i, Tuple(":", ",", "}"), true);
		}
		else if (tokens[i] == ":")
		{
			bool valid = MatchAnyOf(tokens, i, Tuple("{", "["));
			valid |= isTypeOf<string>(tokens[i + 1]);
			valid |= isTypeOf<float>(tokens[i + 1]);
			if (!valid)
			{
				Error("Unexpected symbol");
			}
		}
		else if (tokens[i] == "{")
		{
			bool valid = isTypeOf<string>(tokens[i + 1]);
			if (!valid)
			{
				Error("Unexpected symbol");
			}
		}
		else if (tokens[i] == "}")
		{
			MatchAnyOf(tokens, i, Tuple(",", "}"), true);
		}
		else if (tokens[i] == "[")
		{
			bool valid = MatchAnyOf(tokens, i, Tuple("{", "["));
			valid |= isTypeOf<float>(tokens[i + 1]);
			if (!valid)
			{
				Error("Unexpected symbol");
			}
		}
		else if (tokens[i] == "]")
		{
			MatchAnyOf(tokens, i, Tuple("]", "}", ","), true);
		}
		else if (tokens[i] == ",")
		{
			bool valid = MatchAnyOf(tokens, i, Tuple("{", "["));
			valid |= isTypeOf<string>(tokens[i + 1]);
			valid |= isTypeOf<float>(tokens[i + 1]);
			if (!valid)
			{
				Error("Unexpected symbol");
			}
		}
		else if (isTypeOf<float>(tokens[i]))
		{
			MatchAnyOf(tokens, i, Tuple(",", "]", "}"));
		}
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////
	// LOGIC
	
	stack<_Node*> dfs;
	dfs.push(root);
	bool hasType = false;

	while (!dfs.empty())
	{
		// Pop top of stack
		_Node* tmp = dfs.top();
		dfs.pop();

		if(tmp->GetType() == Type::STRING) 
		{
			Node* node1 = (Node*)tmp;

			if(node1->mLeftSide == "type")
			{
				hasType = true;
				void* str1 = (*tmp).GetData();
				string strAsString = *((string*)str1);
				if(!(strAsString == "Point" || strAsString == "MultiPoint"  || strAsString == "LineString"  || 
					strAsString == "MultiLineString"  || strAsString == "Polygon"  || strAsString == "MultiPolygon" /* || 
					strAsString == "GeometryCollection"  || strAsString == "Feature"  || strAsString == "FeatureCollection"*/))
				{
					Error("Error: Contains invalid type!");
				}

				// Determine type
				GeometryType type = GetGeometryType(strAsString);

				switch (type)
				{
				case GeometryType::GEOMPOINT:
					ValidatePoint(tmp);
					break;

				case GeometryType::MULTIPOINT:
					ValidateMultiPointLineString(tmp, "MultiPoint");
					break;

				case GeometryType::LINESTRING:
					ValidateMultiPointLineString(tmp, "LineString");
					break;

				case GeometryType::MULTILINESTRING:
					ValidateMultiLineString(tmp);
					break;

				case GeometryType::POLYGON:
					ValidatePolygon(tmp);
					break;

				case GeometryType::MULTIPOLYGON:
					ValidateMultiPolygon(tmp);
					break;

				default:
					break;
				}
			}
		}

		// Contiue with children
		for(pair<string, _Node*> n : (*tmp).mElements)
		{
			dfs.push(n.second);
		}
	}
		
	if (!hasType)
	{
		Error("Error: No type");
	}

	cout << "File parsed correctly." << endl << "Press return to exit..." << endl;
	cin.get();

	return 0;

}