#include "xml.h"

namespace PA193XML {
	// Node

	void Node::add_child(Node* child) {
		// push the child at the end of the list
		m_children.push_back(child);
	}

	std::string Node::to_text(size_t indent) {
		std::stringstream output;
		// start the opening tag
		for (size_t i = 0; i < indent; i++) output << " ";
		output << "<" << m_tag->m_name;
		if (!m_tag->m_attributes.empty()) {
			for (attribute_list::iterator it = m_tag->m_attributes.begin(); it != m_tag->m_attributes.end(); ++it) {
				if (std::string::npos == it->second.find("\"")) {
					// if there is no " in value, you can use "
					output << " " << it->first << "=\"" << it->second << "\"";
				}
				else {
					// else use '
					output << " " << it->first << "=\'" << it->second << "\'";
				}
			}
		}
		
		// no children and no content -> empty tag
		if (m_children.empty() && m_content.empty()) {
			output << "/>";
			output << std::endl;
			return output.str();
		}

		// end of opening tag
		output << ">";

		// the content
		if (!m_content.empty()) {
			output << m_content;
		}

		// the children
		if (!m_children.empty()) {
			output << std::endl;
			for (std::list<Node*>::iterator it = m_children.begin(); it != m_children.end(); ++it) {
				output << (*it)->to_text(indent + 4);
			}
			for (size_t i = 0; i < indent; i++) output << " ";
		}

		// closing tag
		output << "</" << m_tag->m_name << ">" << std::endl;
		return output.str();
	}

	// Document

	Document::Document(Node* root, attribute_list attributes) {
		// do not allow a NULL pointer for root
		if (root == NULL) {
			throw 0;
		}
		m_root = root;
		m_attributes = attributes;
	}

	std::string Document::to_text() {
		// if there aren't any "attributes" in the document declaration, do not print any
		if (!m_attributes.empty()) {
			std::stringstream output;
			output << "<?xml";
			for (attribute_list::iterator it = m_attributes.begin(); it != m_attributes.end(); ++it) {
				output << " " << it->first << "=\"" << it->second << "\"";
			}
			output << "?>" << std::endl;
			// concatenate root as text
			output << m_root->to_text();
			return output.str();
		}		

		// the same text as root node
		return m_root->to_text();
	}

	// Parser

	std::string Parser::whitespaces = " \t\f\v\n\r"; // white space characters
	std::string Parser::illegal_characters = "!#$%&'()*+,/;<=>?@[]^`{|}~\\\""; // illegal characters

	std::string Parser::parse_name() {
		// no whitespace allowed between < and name (after name is ok)
		static std::string name_endings(" \t\f\v\n\r>/"); // white spaces or closing or / can end a name
		// find where this name ends
		size_t name_end = m_input->find_first_of(name_endings, m_pos);
		if (name_end == std::string::npos) {
			std::string empty;
			m_error_code |= PARSER_MISSING_NAME;
			return empty;
		}
		size_t len = name_end - m_pos;
		m_pos = name_end;

		return m_input->substr(m_pos - len, len); // return the substring
	}

	void Parser::parse_content(std::string& content_so_far) {
		// skip whitespace before content
		skip_whitespace();
		// search until first opening symbol
		static std::string content_endings("<"); // opening
		size_t content_end = m_input->find_first_of(content_endings, m_pos);
		if (content_end == std::string::npos) {
			return;
		}
		// length of the content
		size_t len = content_end - m_pos;
		// move after the content
		position_move(content_end);
		// extract the content
		std::string new_content = m_input->substr(m_pos - len, len);
		size_t last_non_space = new_content.find_last_not_of(whitespaces);
		if (last_non_space > 0) {
			// erase every trailing space
			new_content.erase(last_non_space + 1);
		}		
		content_so_far.append(new_content); // append the substring
		content_so_far.append(" ");
	}

	attribute_list Parser::parse_attributes() {
		attribute_list attributes;

		// attribute list will end with some of these
		static std::string attribute_name_endings(" \t\f\v\n\r="); // whitespace or =

		// find beginning of the next tag < or end of file
		size_t next_tag_begin = m_input->find_first_of("<", m_pos);
		if (next_tag_begin == std::string::npos) {
			next_tag_begin = m_input->size();
		}

		// generate a substring from m_pos to next_tag_begin
		std::string attributes_text = m_input->substr(m_pos, next_tag_begin - m_pos);

		// find the last closing character of this tag >, that is before this <
		size_t len = attributes_text.find_last_of(">");
		if (len == std::string::npos) {
			// this tag is not closed
			m_error_code |= PARSER_MISSING_GT;
			return attributes;
		}

		size_t attribute_end = len + m_pos;		

		if ((*m_input)[attribute_end - 1] == '/' || (*m_input)[attribute_end - 1] == '?') {
			// if there is />, not only >, then attributes end 1 sooner
			// also forget ? for now
			attribute_end--;
			len--;
		}

		// move pointer to the end of this tag
		m_pos = attribute_end;

		// erase from string everything beginning with /> or >
		attributes_text.erase(len);		
		// erase trailing whitespace
		len = attributes_text.find_last_not_of(whitespaces);
		if (len == std::string::npos) {
			// this string is empty
			return attributes;
		}
		len++; // want to keep the last nonempty char
		// erase other data
		attributes_text.erase(len);

		// now work only with attribute_text
		size_t local_pos = 0;
		while (local_pos < len) {
			// begin to look for attributes after whitespace
			local_pos = attributes_text.find_first_not_of(whitespaces, local_pos);
			if (std::string::npos == local_pos) {
				// only empty spaces left before > or />
				break;
			}
			// find end of attribute name, will end in = or whitespace
			size_t attribute_name_end = attributes_text.find_first_of(attribute_name_endings, local_pos);
			if (attribute_name_end == local_pos) {
				// name has 0 length
				m_error_code |= PARSER_BAD_ATTRIBUTES;
				return attributes;
			}
			// copy the name
			std::string name = attributes_text.substr(local_pos, attribute_name_end - local_pos);
			// check if name contains illegal characters
			if (std::string::npos != name.find_first_of(illegal_characters)) {
				m_error_code |= PARSER_ILLEGAL_CHAR;
				return attributes;
			}
			// move at the end of name
			local_pos = attribute_name_end;
			// look for closest =
			size_t next_del = attributes_text.find_first_of("=", local_pos);
			if (next_del == std::string::npos) {
				m_error_code |= PARSER_BAD_ATTRIBUTES;
				return attributes;
			}
			// start after =
			local_pos = next_del + 1;
			size_t close_del;
			
			if (std::string::npos == (next_del = attributes_text.find_first_of("\"", local_pos))
				|| (next_del > attributes_text.find_first_of("\'", local_pos))) {
				// attributes are not open by " or ' is sooner
				if (std::string::npos == (next_del = attributes_text.find_first_of("\'", local_pos))) {
					// attributes are not open by ' either
					m_error_code |= PARSER_BAD_ATTRIBUTES;
					return attributes;
				}
				else {					
					if (next_del != attributes_text.find_first_not_of(whitespaces, local_pos)) {
						// there is something between = and ' and it is not space
						m_error_code |= PARSER_BAD_ATTRIBUTES;
						return attributes;
					}
					local_pos = next_del + 1;					
					if (std::string::npos == (close_del = attributes_text.find_first_of("\'", local_pos))) {
						// attributes are not closed by '
						m_error_code |= PARSER_BAD_ATTRIBUTES;
						return attributes;
					}
				}
			}
			else {
				// attributes are open by "
				if (next_del != attributes_text.find_first_not_of(whitespaces, local_pos)) {
					// there is something between = and " and it is not space
					m_error_code |= PARSER_BAD_ATTRIBUTES;
					return attributes;
				}
				local_pos = next_del + 1;
				
				if (std::string::npos == (close_del = attributes_text.find_first_of("\"", local_pos))) {
					// attributes are not closed by "
					m_error_code |= PARSER_BAD_ATTRIBUTES;
					return attributes;
				}
			}
			std::string value = attributes_text.substr(local_pos, close_del - local_pos);
			if (close_del == local_pos) {
				// value has 0 length
				m_error_code |= PARSER_BAD_ATTRIBUTES;
				return attributes;
			}
			local_pos = close_del + 1;			

			// put the attributes to map
			// if the key is already used, second part of pair is false
			// this is error - duplicite attribute name
			std::pair<attribute_list::iterator, bool> pair = attributes.insert(std::pair<std::string, std::string>(name, value));
			if (!pair.second) {
				// duplicite attributes
				m_error_code |= PARSER_BAD_ATTRIBUTES;
				return attributes;
			}
		}
		return attributes;
	}

	bool Parser::position_move(size_t new_pos) {
		if (new_pos > m_input->size()) {
			// cannot move there, end of input
			m_error_code |= PARSER_END_INPUT;
			return false;
		}
		m_pos = new_pos;
		return true;
	}

	size_t Parser::skip_whitespace() {
		size_t whitespaces_end = m_input->find_first_not_of(whitespaces, m_pos);
		if (whitespaces_end == std::string::npos) {
			return m_pos;
		}
		m_pos = whitespaces_end;
		return whitespaces_end;
	}

	Tag* Parser::parse_tag(tag_type* type, bool only_comments) {
		if (type == NULL) {
			return NULL;
		}

		*type = TAG_UNKNOWN;

		size_t first_pos = 0;
		size_t first_not_pos;

		bool is_comment = true;

		while (is_comment) {

			first_pos = m_input->find_first_of('<', m_pos); // get position of first <

			if (first_pos == std::string::npos) {
				// missing <

				// if we are only looking for comments, then missing < is not an error
				if (only_comments) {
					return NULL;
				}

				m_error_code |= PARSER_MISSING_LT;
				return NULL;
			}

			first_not_pos = skip_whitespace();

			if (first_pos != first_not_pos) {
				// there is something else before <
				m_error_code |= PARSER_UNEXPECTED_CONTENT;
				return NULL;
			}

			// check if this < is on the same position as closest coment <!--
			if (first_pos == m_input->find("<!--", m_pos)) {
				// this is not a tag, but a comment - delete it

				// safely move after the end of the <!--, so that we can find the ending --
				if (!position_move(m_pos + 4)) {
					return NULL;
				}

				// look for its ending -->, note: -- is not allowed inside of a comment except for the ending
				size_t comment_end = m_input->find("--", m_pos);
				if (comment_end == std::string::npos) {
					// no matching closing string
					m_error_code |= PARSER_BAD_COMMENT;
					return NULL;
				}
				// remove the comment
				// check if can remove -->
				if (comment_end + 2 > m_input->size()) {
					// not enough characters in buffer
					m_error_code |= PARSER_END_INPUT;
					return NULL;
				}

				// check if it really ends with >
				if ((*m_input)[comment_end + 2] != '>') {
					// -- is used somehow else than -->
					m_error_code |= PARSER_END_INPUT;
					return NULL;
				}

				// +3 because including -->
				m_input->erase(first_pos, comment_end - first_pos + 3);

				// move back because we deleted <!--
				position_move(m_pos - 4);
			}
			else {
				if (only_comments) {
					return NULL;
				}
				is_comment = false;
			}

		}
		// advance to next char after <
		if (!position_move(first_pos + 1)) {
			return NULL;
		}

		if ((*m_input)[m_pos] == '/') {
			*type = TAG_CLOSING;
			// move by one
			if (!position_move(m_pos + 1)) {
				return NULL;
			}
		}

		std::string name = parse_name();

		if (name.empty()) {
			m_error_code |= PARSER_MISSING_NAME;
			return NULL;
		}

		size_t illegal_pos = name.find_first_of(illegal_characters);

		// fail if there are any illegal characters in name
		if (std::string::npos != illegal_pos) {
			m_error_code |= PARSER_ILLEGAL_CHAR;
			return NULL;
		}

		// fail if name begins with number, . or -
		illegal_pos = name.find_first_of("0123456789.-");
		if (0 == illegal_pos) {
			m_error_code |= PARSER_ILLEGAL_CHAR;
			return NULL;
		}

		// skip space and parse attributes
		skip_whitespace();
		attribute_list attributes = parse_attributes();
		if (m_error_code != 0) {
			return NULL;
		}

		// closing tags must not have any attribtues
		if (*type == TAG_CLOSING && !attributes.empty()) {
			m_error_code |= PARSER_ATTRIBUTES_IN_CLOSING;
			return NULL;
		}

		// if attribute parsing ended at / and this tag is already closing, then it is error
		if ((*m_input)[m_pos] == '/') {
			if (*type == TAG_CLOSING) {
				m_error_code |= PARSER_END_SLASH_IN_CLOSING | PARSER_BAD_SLASH;
				return NULL;
			}
			*type = TAG_EMPTY;
			// move by one
			// if this tag is trying to be empty, but after / there is nothing
			if (!position_move(m_pos + 1)) {
				return NULL;
			}
		}

		// if there isn't > at the end of this tag
		if ((*m_input)[m_pos] != '>') {
			m_error_code |= PARSER_MISSING_GT;
			return NULL;
		}
		// move pointer after the closing tag >
		if (!position_move(m_pos + 1)) {
			return NULL;
		}
		// if we still don't know what type this is, then it is opening tag
		if (*type == TAG_UNKNOWN) {
			*type = TAG_OPENING;
		}

		return new Tag(name, attributes);
	}

	attribute_list Parser::parse_declaration() {
		// parse the "attributes" in <?xml ... ?> int this
		attribute_list attributes;

		size_t beginning = m_input->find_first_of("<");
		if (beginning == std::string::npos ) {
			// missing <
			m_error_code |= PARSER_MISSING_LT;
			return attributes;
		}
		if (beginning != m_input->find_first_not_of(whitespaces)) {
			// there is something before <
			m_error_code |= PARSER_ILLEGAL_CHAR;
			return attributes;
		}
		if (m_input->size() < beginning + 1) {
			// nothing left to read
			m_error_code |= PARSER_END_INPUT;
			return attributes;
		}
		if ((*m_input)[beginning + 1] != '?') {
			// second character isn't ?
			// this is not error, there just isn't a declaration
			return attributes;
		}
		// there is a declaration and it begins with <?

		// now we require that this the first character of a file
		if (beginning != 0) {
			// nothing left to read
			m_error_code |= PARSER_BAD_DECLARATION;
			return attributes;
		}

		// if we cannot move further, it is error
		// m_pos == beginning + 2
		if (!position_move(beginning + 2)) {
			return attributes;
		}
		if (m_pos + 5 >= m_input->size()) {
			// if there isn't room for xml?>
			m_error_code |= PARSER_END_INPUT;
			return attributes;
		}
		// this should be xml or XML
		std::string xml = m_input->substr(m_pos, 3);
		if (xml.compare("xml") != 0 && xml.compare("XML") != 0) {
			// if there isn't xml directly after <?
			m_error_code |= PARSER_BAD_NODE;
			return attributes;
		}
		m_pos = m_pos + 3;
		// now we expect attribute-like construct, so we use the parse_attributes

		attributes = parse_attributes();
		// m_pos now points at closing ? or /
		if ((*m_input)[m_pos] != '?') {
			// second character from the end isn't ?
			m_error_code |= PARSER_ILLEGAL_CHAR;
			return attributes;
		}
		// move from ? after the >. there must be ?> in sequence because of parse_attributes();
		m_pos += 2;

		// if <?xml is defined, then it must have at least version and it must be first
		if (attributes.empty()) {
			// <?xml... is present, but does not contain version
			m_error_code |= PARSER_BAD_ATTRIBUTES;
			return attributes;
		}

		return attributes;
	}

	Node* Parser::parse_node(Tag* previous_tag) {
		Tag* opening;
		tag_type type = TAG_UNKNOWN;

		if (previous_tag == NULL) {
			size_t first_pos, first_not_pos;
			first_pos = m_input->find_first_of('<', m_pos); // get position of first <

			if (first_pos == std::string::npos) {
				// missing <
				m_error_code |= PARSER_MISSING_LT;
				return NULL;
			}

			first_not_pos = skip_whitespace();

			if (first_pos != first_not_pos) {
				// there is something else before <
				m_error_code |= PARSER_UNEXPECTED_CONTENT;
				return NULL;
			}

			opening = parse_tag(&type);
		}
		else {
			opening = previous_tag;
			type = TAG_OPENING;
		}

		if (opening == NULL) {
			return NULL;
		}

		if (type == TAG_UNKNOWN || type == TAG_CLOSING) {
			delete opening;
			m_error_code |= PARSER_MISMATCHED_TAGS;
			return NULL;
		}

		std::list<Node*> children;
		std::string content;

		// special case = empty root node
		if (type == TAG_EMPTY) {
			return new Node(opening, children);
		}

		Node* this_node = new Node(opening, children);
		parse_content(content);
		this_node->set_content(content);

		// infinite cycle
		for (;;) {
			type = TAG_UNKNOWN;
			Tag* next = parse_tag(&type);

			if (next == NULL) {
				delete this_node;
				m_error_code |= PARSER_MISMATCHED_TAGS;
				return NULL;
			}

			// closing tag
			if (type == TAG_CLOSING) {
				// name matches
				if (opening->m_name.compare(next->m_name) == 0) {
					delete next;
					return this_node;
				}
				// name does not match
				delete next;
				delete this_node;
				m_error_code |= PARSER_MISMATCHED_TAGS;
				return NULL;
			}
			// other type of tag -> children
			if (type == TAG_EMPTY) {
				this_node->add_child(new Node(next));
			}
			if (type == TAG_OPENING) {
				// recursion
				Node* child = parse_node(next);
				if (m_error_code != 0) {
					delete this_node;
					delete child;
					return NULL;
				}
				if (child != NULL) {
					this_node->add_child(child);
				}				
			}
			parse_content(content); // add whatever content there is between tags
			this_node->set_content(content);
		}	
	}

	unsigned int Parser::get_error_code() {
		return m_error_code;
	}

	Document* Parser::parse(std::string* input) {
		m_error_code = 0;
		if (input == NULL || input->empty()) {
			m_error_code |= PARSER_BAD_INPUT;
			return NULL;
		}

		m_input = input;
		m_pos = 0;
		
		// there must be no comments before document declaration
		// get the "attributes" from <?xml ... ?>
		attribute_list attributes = parse_declaration();
		if (m_error_code != 0) {
			m_error_code |= PARSER_BAD_ATTRIBUTES;
			return NULL;
		}
		
		// delete comments at the beginning
		tag_type type = TAG_UNKNOWN;
		parse_tag(&type, true);
		if (m_error_code != 0) {
			m_error_code |= PARSER_BAD_COMMENT;
			return NULL;
		}

		// get the root node
		Node* root = parse_node();
		if (root == NULL || m_error_code != 0) {
			delete root;
			m_error_code |= PARSER_BAD_NODE;
			return NULL;
		}

		// remove comments after root
		parse_tag(&type, true);
		if (m_error_code != 0) {
			m_error_code |= PARSER_BAD_COMMENT;
			return NULL;
		}

		// look if there is trash after ending of root
		size_t extra_char = m_input->find_first_not_of(whitespaces, m_pos);

		if (extra_char != std::string::npos) {
			delete root;
			// some non-empty characters after root element
			m_error_code |= PARSER_BAD_NODE;
			return NULL;
		}

		Document* document = new Document(root, attributes);
		return document;
	}
}
