#ifndef PA193XML_H
#define PA193XML_H

#include <list>
#include <string>
#include <sstream>
#include <algorithm>
#include <map>

namespace PA193XML {
	// "list" of attributes is map of attribute name (key) to attribute value (value)
	typedef std::map<std::string, std::string> attribute_list;

	/**
	 * Class representing a single tag <opening>, </closing> or <empty/>
	 */
	class Tag {
	public:
		std::string m_name; // name of the tag
		attribute_list m_attributes; // list of attribute names and values
	
		/**
		 * Constructor must set the member variables
		 */
		Tag(std::string name, attribute_list attributes) 
			: m_name(name), m_attributes(attributes) {};
	};

	/**
	 * Class representing a node in DOM tree. Has values from opening tag, text content and children.
	 */
	class Node {
	private:
		Tag* m_tag; // tag which contains name and attributes
		std::string m_content; // text content from inbetween tags
		std::list<Node*> m_children; // child nodes
	public:
		/**
		 * Node with only tag, no content or children
		 */
		Node(Tag* tag) : m_tag(tag) {}
		/**
		 * Node with tag, children, but no content
		 */
		Node(Tag* tag, std::list<Node*> children)
			: m_tag(tag), m_children(children) {}
		/**
		 * delete the tag and recursively delete all children
		 */
		~Node() {
			delete m_tag;
			for (std::list<Node*>::iterator it = m_children.begin(); it != m_children.end(); ++it) {
				delete *it;
			}
		}
		/**
		 * Set content
		 */
		void set_content(std::string content) { m_content = content; }
		/**
		 * Add a child node at the end of the list
		 */
		void add_child(Node* child);
		/**
		 * Return text representation of this node. Begin with indentation indent and at every level increase by 4 spaces
		 */
		std::string to_text(size_t indent = 0);
	};

	/**
	 * Class representing a document.
	 */
	class Document {
	private:
		attribute_list m_attributes; // "attributes" from <?xml ... ?>
		Node* m_root; // root node
	public:
		Document(Node* root, attribute_list attributes);
		~Document() {
			delete m_root;
		}
		/**
		 * Return text representation of this document.
		 */
		std::string to_text();
	};

	/**
	 * Represent tag types.
	 */
	typedef enum {
		TAG_UNKNOWN,
		TAG_OPENING, // <name> 
		TAG_CLOSING, // </name>
		TAG_EMPTY // <name/>
	} tag_type;

	/**
	 * Class representing the parser.
	 */
	class Parser {
	private:
		std::string* m_input; // pointer to the string with content
		size_t m_pos; // position in the string
		unsigned int m_error_code; // error state of the parser, 0 means no error

		static std::string whitespaces; // white space characters
		static std::string illegal_characters; // illegal characters
		/**
		 * Return name at the current position, empty string if name cannot be parsed. Also sets error code on failure.
		 */
		std::string parse_name();
		/**
		 * Concatenates content on the current position to the content_so_far.
		 */
		void parse_content(std::string& content_so_far);
		/**
		 * Return attributes at the current position, empty list if attributes cannot be parsed. Also sets error code on failure.
		 */
		attribute_list parse_attributes();
		/**
		 * Return pointer to dynamically allocated Tag, NULL if tag cannot be parsed. Also sets error code on failure.
		 * if only_comments == true, do not throw an error if you don't find a tag and do not parse any tags
		 */
		Tag* parse_tag(tag_type* type, bool only_comments = false);
		/**
		 * Move to given position and return true, false if the new position is after buffer. Also sets error code on failure.
		 */
		bool position_move(size_t new_pos);
		/**
		 * Advacnce m_pos to the next position, which is not whitespace.
		 */
		size_t skip_whitespace();
		/**
		 * Removes all correctly matching comments <!-- -->. Also sets error code on failure.
		 */
		attribute_list parse_declaration();
		/**
		 * Return pointer to dynamically allocated Node. If a previous_tag is supplied, does not parse for tag. Returns NULL and also sets error code on failure.
		 */
		Node* parse_node(Tag* previous_tag = NULL);
	public:
		Parser() : m_input(NULL), m_pos(0), m_error_code(0) {};
		/**
		 * Return the error code.
		 */
		unsigned int get_error_code();
		/**
		 * Return pointer to dynamically allocated Document, if input contains valid document.
		 * Return NULL if document cannot be parsed. Also sets error code on failure.
		 */
		Document* parse(std::string* input);
	};

	/**
	 * Representation of error codes. Error state can be logical and of any subset of error codes.
	 */
	typedef enum {
		PARSER_BAD_INPUT = 1,
		PARSER_END_INPUT = 2,
		PARSER_BAD_NODE = 4,
		PARSER_UNEXPECTED_CONTENT = 8,
		PARSER_MISSING_LT = 16,
		PARSER_MISSING_GT = 32,
		PARSER_BAD_SLASH = 64,
		PARSER_BAD_CLOSING = 128,
		PARSER_MISSING_NAME = 256,
		PARSER_ATTRIBUTES_IN_CLOSING = 512,
		PARSER_END_SLASH_IN_CLOSING = 1024,
		PARSER_MISMATCHED_TAGS = 2048,
		PARSER_BAD_ATTRIBUTES = 4096,
		PARSER_ILLEGAL_CHAR = 8192,
		PARSER_BAD_COMMENT = 2 * 8192,
		PARSER_BAD_DECLARATION = 4 * 8192
	} error_codes;	
}
#endif // PA193XML_H