/*
 ============================================================================
 Name        : iso2709.c
 Description : ISO 2709 parser
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <regex.h>
#include <string.h>
#include <ctype.h>
#include "iso2709.h"

const unsigned int int_chunk_size = 10 * 1024 * sizeof(char);
const unsigned int int_max_record_size = 1024 * 100 * sizeof(char);

const unsigned int marc_header_length = 24;
const unsigned int marc_record_splitter = 0x1d;
const unsigned int marc_end_field = 0x1e;
const unsigned int marc_subfield_indicator = 0x1f;

unsigned long long parsed_records = 0;

regex_t regex_numeric_only;
regex_t regex_marc_endfield;

int _iso2709_init() {
	int regex_init = 0;
	regex_init = regcomp(&regex_numeric_only, "^[0-9]+$", REG_EXTENDED);
	if (regex_init) {
		return 1;
	}

	regex_init = regcomp(&regex_marc_endfield, "\x1e", 0);
	if (regex_init) {
		return 1;
	}
	return 0;
}

void _iso2709_free() {
	regfree(&regex_marc_endfield);
	regfree(&regex_numeric_only);
}

int _iso2709_check_numbers(const char* in_str) {
	return regexec(&regex_numeric_only, in_str, 0, NULL, 0);
}

size_t _iso2709_read_next_record(FILE* in_file, char** outptr_bufffer) {
	if (!in_file)
		return 0;

	size_t written_count = 0;
	size_t read = 0;
	char* record_buffer = NULL;
	char* temp_buffer = NULL;

	do {
		temp_buffer = (char*) realloc(record_buffer,
				written_count * sizeof(char) + int_chunk_size);
		//check allocation
		if (!temp_buffer) {
			if (record_buffer) {
				free(record_buffer);
			}
			return 0;
		} else {
			record_buffer = temp_buffer;
		}

		//read another chunk of file
		read = fread(record_buffer + written_count, sizeof(char),
				int_chunk_size, in_file);
		if (read == 0) {
			free(record_buffer);
			return 0;
		}

		for (size_t i = written_count; i < written_count + read; i++) {
			//search for end of record
			if (record_buffer[i] == (signed) marc_record_splitter) {
				*outptr_bufffer = record_buffer;
				//rewind file pointer to the begin of next record
				for (size_t j = read - (i % int_chunk_size) - 1; j > 0; j--) {
					fseek(in_file, -1, SEEK_CUR);
				}
				return i + 1;
			}

		}
		written_count += read;
	} while (written_count < int_max_record_size);

	if (record_buffer)
		free(record_buffer);
	return 0;
}

/**
 * read data offset from record
 * @return data offset of 0
 */
size_t _iso2709_get_data_offset(const char* in_record,
		const size_t in_record_length) {
	if (in_record_length <= marc_header_length) {
		return 0;
	}
	char header_str[marc_header_length + 1];
	memset(header_str, 0, sizeof(char) * (marc_header_length + 1));
	strncpy(header_str, in_record, marc_header_length);

	char offset[6];
	memset(offset, 0, 6 * sizeof(char));
	strncpy(offset, header_str + 12, 5);
	if (_iso2709_check_numbers(offset) != 0) {
		return 0;
	}

	return strtoul(offset, NULL, 10);
}

int _iso2709_print_header(const char* in_record, const size_t in_record_length,
		FILE* out_file) {
	if (in_record_length <= marc_header_length) {
		return 1;
	}
	char header[marc_header_length + 1];
	memset(header, 0, marc_header_length * sizeof(char) + 1);
	strncpy(header, in_record, marc_header_length);
	int result = fprintf(out_file, "LDR  %s\n", header);
	return result < 0 ? 1 : 0;
}

void _iso2709_replace_subfield_indicators(char* in_content, size_t length) {
	for (unsigned int i = 0; i < length; i++) {
		if (in_content[i] == (signed) marc_subfield_indicator) {
			in_content[i] = '$';
		}
	}
}

int _iso2709_parse_record(const char* in_record, const size_t in_length,
		FILE* out_file) {
	if (!out_file) {
		return 9;
	}

	if (!in_record)
		return 1;
	//record too short
	if (in_length <= marc_header_length)
		return 2;
	//check for proper end of record
	if (in_record[in_length - 1] != (signed) marc_record_splitter)
		return 3;

	size_t offset = _iso2709_get_data_offset(in_record, in_length);
	if (offset == 0)
		return 4;

	//ivalid padding
	if ((offset - 1) % 12 != 0)
		return 5;

	size_t current_data_offset = 0;
	size_t field_count = ((offset - 1) / 12) - 2;
	if (field_count < 2) {
		return 6;
	}

	//first write to stream, perform check
	if (_iso2709_print_header(in_record, int_max_record_size, out_file) != 0) {
		return 9;
	}


	//parse each field
	for (unsigned int i = 0; i < field_count; i++) {
		char field_spec[13];
		memset(field_spec, 0, 13 * sizeof(char));
		strncpy(field_spec, in_record + marc_header_length + (i * 12), 12);

		if (_iso2709_check_numbers(field_spec) != 0) {
			return 7;
		}

		char field_name[4];
		memset(field_name, 0, 4 * sizeof(char));
		strncpy(field_name, field_spec, 3);

		char field_length_str[5];
		memset(field_length_str, 0, 5 * sizeof(char));
		strncpy(field_length_str, field_spec + 3, 4);
		size_t field_length = strtoul(field_length_str, NULL, 10);

		char field_offset_str[6];
		memset(field_offset_str, 0, 6 * sizeof(char));
		strncpy(field_offset_str, field_spec + 7, 5);
		size_t field_offset = strtoul(field_offset_str, NULL, 10);

		current_data_offset = offset + field_offset;
		//are we still valid?
		if (current_data_offset + field_length >= in_length
				|| in_record[current_data_offset - 1]
						!= (signed) marc_end_field) {
			return 8;
		}

		char field_content_str[field_length];
		memset(field_content_str, 0, field_length * sizeof(char));
		strncpy(field_content_str, in_record + current_data_offset,
				field_length - 1);
		fprintf(out_file, "%s", field_name);

		//fields 00* shuldn't have indicators
		if (field_name[0] == '0' && field_name[1] == '0') {
			fprintf(out_file, "  ");
		}

		_iso2709_replace_subfield_indicators(field_content_str,
				field_length - 1);
		fprintf(out_file, "%s\n", field_content_str);
	}
	return 0;
}

int iso2709_parse_file(FILE* in_file, FILE* out_file) {
	if (!in_file || !out_file) {
		return 1;
	}

	if (_iso2709_init() != 0) {
		fprintf(stderr, "ISO2709 init failed");
		return 1;
	}

	size_t lenght = 0;

	do {
		char* record = NULL;
		char* record_ltrim = NULL;
		lenght = _iso2709_read_next_record(in_file, &record);
		if (lenght == 0) {
			if (record) {
				free(record);
			}
			_iso2709_free();
			return parsed_records > 0 ? 0 : 1;
		}
		if (!record) {
			_iso2709_free();
			return parsed_records > 0 ? 0 : 1;
		}

		//skip possible white space before record
		record_ltrim = record;
		size_t i = 0;
		for (i = 0; i < lenght; i++) {
			if (isspace(record[i])) {
				record_ltrim++;
			} else {
				break;
			}
		}

		int result = _iso2709_parse_record(record_ltrim, lenght - i, out_file);

		free(record);
		if (result != 0) {
			_iso2709_free();
			return result;
		}
		parsed_records++;
	} while (lenght > 0);

	_iso2709_free();
	return parsed_records > 0 ? 0 : 1;
}

