
#include <cstdint>
#include <cassert>
#include <stdexcept>
#include <vector>
#include <string>
#include <sstream>

#include <sexpressions.hpp>

namespace sexp {

static int
_is_whitespace(int c)
{
	return c == ' ' || c == '\t' || c == '\v' || c == '\f' || c == '\n' || c == '\r';
}

static size_t
_check_length(uint64_t l)
{
	if ((uint64_t)(size_t)l != l) {
		throw new std::runtime_error("Length too long");
	}
	
	return (size_t)l;
}

static bool
_is_valid_token_char(int c)
{
	return isalnum(c) ||
		c == '-' || c == '.' || c == '/' || c == '_' ||
		c == ':' || c == '*' || c == '+' || c == '=';
}

static bool
_is_base64_digit(int c)
{
	return isalnum(c) || c == '/' || c == '+';
}

static int
_hex_num(int c)
{
	if (isdigit(c)) {
		return c - '0';
	}
	if (islower(c)) {
		return c + 10 - 'a';
	}
	if (isupper(c)) {
		return c + 10 - 'A';
	}
	return 0;
}

void
Parser::_advance()
{
	if (_current < 0) {
		_current = _input->get();
	}
	_current = _input->get();
}

bool
Parser::eof()
{
	_skip_space();
	return _input->eof();
}

int
Parser::_current_byte()
{
	if (_current < 0) {
		_current = _input->get();
	}
	
	return _current;
}

int
Parser::_get_byte()
{
	int c = _current_byte();
	_advance();
	return c;
}

void
Parser::_skip_space()
{
	while (_is_whitespace(_current_byte())) {
		_advance();
	}
}

uint64_t
Parser::_parse_decimal()
{
	// Check that there is at least one digit (should be ensured by caller).
	assert(isdigit(_current_byte()));
	
	// Check that there are no unnecessary leading zeros.
	if (_pedantic && _current_byte() == '0' && (_advance(), true) && isdigit(_current_byte())) {
		throw new std::runtime_error("Unnecessary leading zero in an octet string length.");
	}
	
	// Read the digits.
	
	const uint64_t MAX_SAFE_NUMBER[] = { 1844674407370955161, 1844674407370955161,
		1844674407370955161, 1844674407370955161, 1844674407370955161, 1844674407370955161,
		1844674407370955160, 1844674407370955160, 1844674407370955160, 1844674407370955160 };
	
	uint64_t result = 0;
	
	while (isdigit(_current_byte())) {
		int digit = _current_byte() - '0';
		_advance();
		
		if (result > MAX_SAFE_NUMBER[digit]) {
			throw new std::runtime_error("Octet string length too large.");
		}
		
		result = 10 * result + digit;
	}
	
	return result;
}

Expression *
Parser::_parse_string()
{
	std::vector<uint8_t> hint;
	std::vector<uint8_t> data;
	
	_skip_space();
	if (_current_byte() == '[') {
		_advance();
		_parse_string_raw(hint);
		_skip_space();
		
		if (_current_byte() != ']') {
			throw new std::runtime_error("Malformed display hint.");
		}
		_advance();
	}
	
	_parse_string_raw(data);
		
	Expression *expr = new Expression();
	expr->is_octets = true;
	expr->display_hint = std::move(hint);
	expr->octets = std::move(data);
	return expr;
}

void
Parser::_parse_string_raw(std::vector<uint8_t> &data)
{
	bool has_length = false;
	uint64_t length = 0;
	
	_skip_space();
	
	if (isdigit(_current_byte())) {
		length = _parse_decimal();
		has_length = true;
	}
	
	if (has_length) {
		data.reserve(_check_length(length));
	}
	
	int c = _current_byte();
	
	switch (c) {
	case ':':
		if (!has_length) {
			throw new std::runtime_error("Verbatim string must be preceded by a length.");
		}
		_parse_verbatim(data, length);
		break;
	case '"':
		_parse_quoted(data);
		break;
	case '#':
		_parse_hex(data);
		break;
	case '|':
		_parse_base64(data);
		break;
	default:
		if (!has_length && _is_valid_token_char(c) && !isdigit(c)) {
			_parse_token(data);
			break;
		}
		throw new std::runtime_error("Invalid character or end of file.");
	}
	
	/* Check size. */
	
	if (_pedantic && has_length && data.size() != length) {
		throw new std::runtime_error("Announced octet string size is not the same as real size.\n");
	}
	
	_skip_space();
}

void
Parser::_parse_verbatim(std::vector<uint8_t> &data, uint64_t length)
{
	assert(_current_byte() == ':');
	_advance();
	
	for (; length > 0; length--) {
		data.push_back(_current_byte());
		_advance();
	}
}

void
Parser::_parse_quoted(std::vector<uint8_t> &data)
{
	assert(_current_byte() == '"');
	_advance();
	
	int c;
	
	while ((c = _current_byte()) != '"') {
		_advance();
		
		if (c != '\\') {
			data.push_back(c);
			continue;
		}
		
		int hi, lo, mid, next;
		
		switch (c = _get_byte()) {
		case 'b': data.push_back('\b'); continue;
		case 't': data.push_back('\t'); continue;
		case 'v': data.push_back('\v'); continue;
		case 'n': data.push_back('\n'); continue;
		case 'f': data.push_back('\f'); continue;
		case 'r': data.push_back('\r'); continue;
		case '"': data.push_back('"');  continue;
		case '\'': data.push_back('\''); continue;
		case '\\': data.push_back('\\'); continue;
		case 'x':
			hi = _get_byte();
			lo = _get_byte();
		
			if (!isxdigit(hi) || !isxdigit(lo)) {
				throw new std::runtime_error("Invalid escape sequence.");
			}
			
			data.push_back(_hex_num(hi) << 4 | _hex_num(lo));
			continue;
		
		case '\n':
		case '\r':
			next = _current_byte();
			if ((next == '\n' || next == '\r') && (next != c)) {
				_advance();
			}
			continue;

		default:
			/* The only other possibility is octal code. */
			
			hi = _get_byte();
			mid = _get_byte();
			lo = _get_byte();
		
			if (hi < '0' || hi > '3' || mid < '0' || mid > '7' || lo < '0' || lo > '7') {
				throw new std::runtime_error("Invalid escape sequence.");
			}
			
			data.push_back((hi - '0') << 6 | (mid - '0') << 3 | (lo - '0'));
			continue;
		}
	}
	
	assert(_current_byte() == '"');
	_advance();
}

void
Parser::_parse_token(std::vector<uint8_t> &data)
{
	assert(!isdigit(_current_byte()));
	assert(_is_valid_token_char(_current_byte()));
	
	while (_is_valid_token_char(_current_byte())) {
		data.push_back(_current_byte());
		_advance();
	}
}

void
Parser::_parse_hex(std::vector<uint8_t> &data)
{
	assert(_current_byte() == '#');
	_advance();
	
	int partial = -1;
	
	while (true) {
		int c = _get_byte();
		
		if (c == '#') {
			break;
		}
		
		if (_is_whitespace(c)) {
			continue;
		}
		
		if (!isxdigit(c)) {
			throw new std::runtime_error("Hexadecimal octet string contains bad characters.\n");
		}
		
		if (partial == -1) {
			partial = c;
		} else {
			uint8_t byte = _hex_num(partial) << 4 | _hex_num(c);
			data.push_back(byte);
			partial = -1;
		}
	}
	
	if (partial != -1) {
		throw new std::runtime_error("Hexadecimal octet string ends with a partial byte.\n");
	}
}

static uint8_t
_b64_decode(int c)
{
	if (c >= 'A' && c <= 'Z') {
		return c - 'A';
	}
	if (c >= 'a' && c <= 'z') {
		return c - 'a' + 26;
	}
	if (c >= '0' && c <= '9') {
		return c - '0' + 52;
	}
	if (c == '+') {
		return 62;
	}
	if (c == '/') {
		return 63;
	}
	
	throw new std::runtime_error("Base64 string contains non-base64 characters.");
}

/**
 * Reads octet string encoded using base64.
 * The current position in the input is expected to be the first character of
 * the base64 representation, and the method interprets characters as base64
 * encoding until
 * 	a) An 'equal' sign is encountered, indicating end of encoding.
 * 	b) A character is encountered that is not in the base64 character set.
 * 
 * After method exists, the current position is the first byte after the
 * decoded base64 representation. 
 */
void
Parser::_read_base64(std::vector<uint8_t> &data)
{
	uint32_t buffer = 0;
	int buffer_fill = 0;
	int eq_signs = 0;
	
	for (;; _advance()) {
		int c = _current_byte();
		
		if (c == '=') {
			_advance();
			eq_signs++;
			while ((c = _current_byte()) == '=') {
				_advance();
				eq_signs++;
			}
			
			break;
		}
		
		if (_is_whitespace(c)) {
			continue;
		}
		
		if (!_is_base64_digit(c)) {
			break;
		}
		
		buffer <<= 6;
		buffer |= _b64_decode(c);
		buffer_fill++;
		
		if (buffer_fill == 4) {
			data.push_back(buffer >> 16);
			data.push_back((buffer >> 8) & 0xff);
			data.push_back(buffer & 0xff);
			
			buffer = 0;
			buffer_fill = 0;
		}
	}
	
	int expected_eqs = 0;
	
	switch (buffer_fill) {
	case 0:
		expected_eqs = 0;
		break;
	case 1:
		throw new std::runtime_error("Invalid Base64 encoding (1).\n");
	case 2:
		expected_eqs = 2;
		data.push_back(buffer >> 4);
		break;
	case 3:
		expected_eqs = 1;
		data.push_back((buffer >> 10) & 0xff);
		data.push_back((buffer >> 2) & 0xff);
		break;
	}
	
	if (eq_signs != 0 && eq_signs != expected_eqs) {
		throw new std::runtime_error("Invalid Base64 encoding (2).\n");
	}
}

void
Parser::_parse_base64(std::vector<uint8_t> &data)
{
	assert(_current_byte() == '|');
	_advance();
	
	_read_base64(data);
	_skip_space();
	
	if (_current_byte() != '|') {
		throw new std::runtime_error("Expected '|' after base64 octet string.");
	}
	_advance();
}

Expression *
Parser::_parse_list()
{
	_advance();
	_skip_space();
	
	// TODO: Properly dispose of allocated expressions in case of exception.
	std::vector<Expression*> children;
	
	while (_current_byte() != ')') {
		if (eof()) {
			throw new std::runtime_error("Unexpected end of file.");
		}
		
		Expression *ex = parse_expression();
		children.push_back(ex);
		_skip_space();
	}
	
	_advance();
	_skip_space();
	
	Expression *expr = new Expression();
	expr->is_octets = false;
	expr->children = std::move(children);
	return expr;
}

Expression *
Parser::parse_expression()
{
	_skip_space();
	
	if (_current_byte() == '(') {
		return _parse_list();
	}
	
	if (_current_byte() == '{') {
		_advance();
		_skip_space();
		std::vector<uint8_t> data;
		_read_base64(data);
		_skip_space();
		if (_current_byte() != '}') {
			throw new std::runtime_error("Expected '}' after base64 encoded expression.");
		}
		_advance();
		
		// Instantiate a new parser on the decoded string.
		std::istringstream in(std::string(data.begin(), data.end()));
		Parser p(in);
		return p.parse_expression();
	}
	
	return _parse_string();
}

void
CanonicalWriter::_write_octets(std::vector<uint8_t> &data)
{
	*_output << data.size() << ":";
	for (int i = 0; i < data.size(); i++) {
		char c = data[i];
		_output->write(&c, 1);
	}
}

void
CanonicalWriter::write(Expression *exp)
{
	if (exp->is_octets) {
		if (exp->display_hint.size() > 0) {
			*_output << "[";
			_write_octets(exp->display_hint);			
			*_output << "]";
		}
		_write_octets(exp->octets);
	} else {
		*_output << "(";
		for (int i = 0; i < exp->children.size(); i++) {
			write(exp->children[i]);
		}
		*_output << ")";
	}
}

}
